// dataset.cpp: implementation of
//		class Value
//		typedef HValue
//		class DataSet
//		typedef HDataSet
//		class Sequence
//////////////////////////////////////////////////////////////////////
/****************************************************************************
          Copyright (C) 1998, Karl Franzens University, Graz

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND THE KARL FRANZENS
          UNIVERSITY DOES NOT MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          owned by the Karl Franzens University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include this copyright notice.
****************************************************************************/

/***************************************************************************
 *
 * Karl Franzens University, Graz
 * UCDMC99 DICOM Network Transport Libraries 
 * Version 0.1 Beta
 *
 * Technical Contact: gunter.zeilinger@kfunigraz.ac.at
 *
 ***************************************************************************/

#include "base.hpp"
#include "DataSet.h"
#include "Strtok.h"

#include "Endian.h"

//////////////////////////////////////////////////////////////////////
// class Value  
//////////////////////////////////////////////////////////////////////

Value::Value(VR vr)
{
	m_vr = vr;
	m_length = 0;
	m_data_p = 0;
	m_seq_p = 0;
}


/*
	Note that the definition for the destructor of Value implies that
	when a Value object is constructed from a void pointer or a Sequence*, 
	it assumes ownership of the pointed-to object, and subsequently deletes
	it the Value object is destroyed.  This behaviour is not documented, no 
	indication is given to the user that this will happen.  Double-free situations
	and their associated problems are likely to occur.  This issue must be addressed.
	
	nb: this is an issue of a POLICY (in this case, ownership) being ill defined or
	unclearly communicated.
	Trevor Morgan. (morgan@sten.sunnybrook.utoronto.ca)
*/

Value::~Value()
{
	delete[] m_data_p;
	delete m_seq_p;
}
/*
	see above comments.  This should use memcpy rather
	than just copying the pointer.
*/
Value::Value(VR vr, UINT32 length, void *data_p)
{
	m_vr = vr;
	m_length = length;
	m_data_p = (char*)data_p;
	m_seq_p = 0;
}

#if 0
//should be:
Value::value(VR vr, UINT32 length,void * data) : 
	m_vr(vr),m_length(length),m_seq_p(NULL)
{
	std::copy(data,m_data_p,length);
}
//or even
template<bool OwnPointer>
class ValueBase{
//...
	Value(VR vr,UINT32 length,void * data) : 
	m_vr(vr),m_length(length),m_seq_p(NULL)
	{
		if(OwnPointer)
			m_data_p=data;
		else
			std::copy(data,m_data_p,length);
	}
//...
};
typedef ValueBase<true> Value;
//i.e, making ownership an explicitly stated policy...

#endif

Value::Value(VR vr, const std::string &str)
{
	m_vr = vr;
	UINT32 len = str.length();
	m_length = (len+1)&~1;
	m_data_p = 0;
	m_seq_p = 0;
	if (m_length > 0)
	{	
		m_data_p = new char[m_length];
	#ifdef WIN32
		str.copy(m_data_p, len);
	#else
		((std::string&)str).copy(m_data_p, len);
	#endif
		if (len&1)
		{
			switch (vr)
			{
			case VR_OB:
			case VR_UN:
			case VR_UI:
				m_data_p[len] = '\0';
				break;
			default:
				m_data_p[len] = ' ';
			}
		}
	}
}

Value::Value(Sequence *seq_p)
{
	m_vr = VR_SQ;
	m_length = seq_p->empty() ? 0 : 0xffffffff;
	m_data_p = 0;
	m_seq_p = seq_p;
}

Value::Value(UINT16 val)
{
	m_vr = VR_US;
	m_length = 2;
	m_data_p = new char[2];
	*((LE_UINT16*)m_data_p) = val;
	m_seq_p = 0;
}

Value::Value(INT16 val)
{
	m_vr = VR_SS;
	m_length = 2;
	m_data_p = new char[2];
	*((LE_INT16*)m_data_p) = val;
	m_seq_p = 0;
}

Value::Value(UINT32 val)
{
	m_vr = VR_UL;
	m_length = 4;
	m_data_p = new char[4];
	*((LE_UINT32*)m_data_p) = val;
	m_seq_p = 0;
}

Value::Value(INT32 val)
{
	m_vr = VR_SL;
	m_length = 4;
	m_data_p = new char[4];
	*((LE_INT32*)m_data_p) = val;
	m_seq_p = 0;
}

Value::Value(float val)
{
	ASSERT(sizeof(float) == 4);
	m_vr = VR_FL;
	m_length = 4;
	m_data_p = new char[4];
	*((LE_float*)m_data_p) = val;
	m_seq_p = 0;
}

Value::Value(double val)
{
	/*
		compile-time asserts would be better here.
	*/
	ASSERT(sizeof(double) == 8);
	m_vr = VR_FD;
	m_length = 8;
	m_data_p = new char[8];
	*((LE_double*)m_data_p) = val;
	m_seq_p = 0;
}

void Value::setData(VR vr, UINT32 length, char* data_p)
{
	ASSERT(m_data_p == 0);
	ASSERT(m_seq_p == 0);
	m_vr = vr;
	m_length = length;
	m_data_p = data_p;
}

void Value::setSQ(Sequence* seq_p)
{
	ASSERT(m_data_p == 0);
	ASSERT(m_seq_p == 0);
	m_vr = VR_SQ;
	m_length = (UINT32)-1;
	m_seq_p = seq_p;
}

std::string Value::str(VR vr) const
{
	const char* left = m_data_p;
	UINT32 len = m_length;
	if (len == 0)
		return std::string();

	switch (vr)
	{
	case VR_UI:
		while (len > 0 && left[len-1] == '\0')
		{
			--len;
		}
		break;
	case VR_AE:
	case VR_CS:
	case VR_DS:
	case VR_IS:
	case VR_LO:
	case VR_SH:
		while (len > 0 && *left == ' ')
		{
			++left;
			--len;
		}
	case VR_DA:
	case VR_LT:
	case VR_PN:
	case VR_ST:
	case VR_TM:
		while (len > 0 && left[len-1] == ' ')
		{
			--len;
		}
	}
	return std::string(left, len);
}

std::string Value::str(VR vr, int index) const
{
	return StrTok(str(vr), '\\')[index];
}

UINT16 Value::getUS(int index) const
{
	if ((index+1)*sizeof(UINT16) > m_length)
		 return 0;
	return ((LE_UINT16*)m_data_p)[index];
}

INT16 Value::getSS(int index) const
{
	if ((index+1)*sizeof(INT16) > m_length)
		 return 0;
	return ((LE_INT16*)m_data_p)[index];
}

UINT32 Value::getUL(int index) const
{
	if ((index+1)*sizeof(UINT32) > m_length)
		 return 0;
	return ((LE_UINT32*)m_data_p)[index];
}

INT32 Value::getSL(int index) const
{
	if ((index+1)*sizeof(INT32) > m_length)
		 return 0;
	return ((LE_INT32*)m_data_p)[index];
}

float Value::getFL(int index) const
{
	if ((index+1)*sizeof(float) > m_length)
		 return 0;
	return ((LE_float*)m_data_p)[index];
}

double Value::getFD(int index) const
{
	if ((index+1)*sizeof(double) > m_length)
		 return 0;
	return ((LE_double*)m_data_p)[index];
}

bool Value::getSafeUS(UINT16* val_p, int index) const
{
	if ((index+1)*sizeof(UINT16) > m_length)
		return false;
	*val_p = ((LE_UINT16*)m_data_p)[index];
	return true;
}

bool Value::getSafeSS(INT16* val_p, int index) const
{
	if ((index+1)*sizeof(INT16) > m_length)
		return false;
	*val_p = ((LE_INT16*)m_data_p)[index];
	return true;
}

bool Value::getSafeUL(UINT32* val_p, int index) const
{
	if ((index+1)*sizeof(UINT32) > m_length)
		return false;
	*val_p = ((LE_UINT32*)m_data_p)[index];
	return true;
}

bool Value::getSafeSL(INT32* val_p, int index) const
{
	if ((index+1)*sizeof(INT32) > m_length)
		return false;
	*val_p = ((LE_INT32*)m_data_p)[index];
	return true;
}

bool Value::getSafeFL(float* val_p, int index) const
{
	if ((index+1)*sizeof(float) > m_length)
		return false;
	*val_p = ((LE_float*)m_data_p)[index];
	return true;
}

bool Value::getSafeFD(double* val_p, int index) const
{
	if ((index+1)*sizeof(double) > m_length)
		return false;
	*val_p = ((LE_double*)m_data_p)[index];
	return true;
}

//////////////////////////////////////////////////////////////////////
// class DataSet : public map<Tag, HValue>
//////////////////////////////////////////////////////////////////////

DataSet::DataSet()
{
}

std::pair<DataSet::iterator, bool> DataSet::insertElem(Tag tag, HValue val_h)
{
	return insert(value_type(tag, val_h));
}

/*
	DataSet is a map of Tag, Handle<Value> pairs.  
	Handle takes care of deleting the Value object that is
	created in the following functions.
*/
std::pair<DataSet::iterator, bool> DataSet::insertElem(Tag tag, VR vr)
{
	return insert(value_type(tag, new Value(vr)));
}

std::pair<DataSet::iterator, bool> DataSet::insertElem(Tag tag, VR vr, UINT32 length, void* data_p)
{
	return insert(value_type(tag, new Value(vr, length, data_p)));
}

std::pair<DataSet::iterator, bool> DataSet::insertElem(Tag tag, Sequence* seq_p)
{
	return insert(value_type(tag, new Value(seq_p)));
}

std::pair<DataSet::iterator, bool> DataSet::insertElem(Tag tag, VR vr, const std::string& val)
{
	return insert(value_type(tag, new Value(vr, val)));
}

std::pair<DataSet::iterator, bool> DataSet::insertElem(Tag tag, UINT16 val)
{
	return insert(value_type(tag, new Value(val)));
}

std::pair<DataSet::iterator, bool> DataSet::insertElem(Tag tag, INT16 val)
{
	return insert(value_type(tag, new Value(val)));
}

std::pair<DataSet::iterator, bool> DataSet::insertElem(Tag tag, UINT32 val)
{
	return insert(value_type(tag, new Value(val)));
}

std::pair<DataSet::iterator, bool> DataSet::insertElem(Tag tag, INT32 val)
{
	return insert(value_type(tag, new Value(val)));
}

std::pair<DataSet::iterator, bool> DataSet::insertElem(Tag tag, float val)
{
	return insert(value_type(tag, new Value(val)));
}

std::pair<DataSet::iterator, bool> DataSet::insertElem(Tag tag, double val)
{
	return insert(value_type(tag, new Value(val)));
}

void DataSet::insertElemFrom(Tag* tags, const DataSet& src)
{
   HValue val_h;
   for (; *tags != TAG_NULL; ++tags)
      if (src.getSafeHValue(val_h, *tags))
         insertElem(*tags, val_h);
}

int DataSet::getTotElem(bool nested) const
{
	int num = size();
	if (nested)
	{
		DataSet::const_iterator ds_end = end();
		DataSet::const_iterator ds_iter = begin();
		while (ds_iter != ds_end)
		{
			const Sequence* pSeq = (ds_iter++)->second->seq();
			if (pSeq)
			{
				Sequence::const_iterator sq_end = pSeq->end();
				Sequence::const_iterator sq_iter = pSeq->begin();
				while (sq_iter != sq_end)
				{
					num += (*(sq_iter++))->getTotElem(nested);
				}
			}
		}
	}
	return num;
}

bool DataSet::exist(Tag tag) const
{
	return find(tag) != end();
}

HValue DataSet::getHValue(Tag tag) const
{
   HValue hVal;
   getSafeHValue(hVal, tag);
   return hVal;
}

const void* DataSet::getData(Tag tag) const
{
   const void* pData = 0;
   getSafeData(&pData, tag);
   return pData;
}

const Sequence* DataSet::getSQ(Tag tag) const
{
   const Sequence* pSeq = 0;
   getSafeSQ(&pSeq, tag);
   return pSeq;
}

std::string DataSet::getStr(Tag tag, VR vr) const
{
   std::string str;
   getSafeStr(&str, tag, vr);
   return str;
}

std::string DataSet::getStr(Tag tag, VR vr, int index) const
{
   std::string str;
   getSafeStr(&str, tag, vr, index);
   return str;
}

UINT16 DataSet::getUS(Tag tag, int index) const
{
   UINT16 val = 0;
   getSafeUS(&val, tag, index);
   return val;
}

INT16 DataSet::getSS(Tag tag, int index) const
{
   INT16 val = 0;
   getSafeSS(&val, tag, index);
   return val;
}

UINT32 DataSet::getUL(Tag tag, int index) const
{
   UINT32 val = 0;
   getSafeUL(&val, tag, index);
   return val;
}

INT32 DataSet::getSL(Tag tag, int index) const
{
   INT32 val = 0;
   getSafeSL(&val, tag, index);
   return val;
}

float DataSet::getFL(Tag tag, int index) const
{
   float val = 0;
   getSafeFL(&val, tag, index);
   return val;
}

double DataSet::getFD(Tag tag, int index) const
{
   double val = 0;
   getSafeFD(&val, tag, index);
   return val;
}

bool DataSet::getSafeHValue(HValue& val_h, Tag tag) const
{
	const_iterator it = find(tag);
	if (it == end())
		return false;
	val_h = it->second;
	return true;
}

bool DataSet::getSafeData(const void** val_p, Tag tag) const
{
	const_iterator it = find(tag);
	if (it == end())
		return false;
	*val_p = it->second->data();
	return true;
}

bool DataSet::getSafeSQ(const Sequence** val_p, Tag tag) const
{
	const_iterator it = find(tag);
	if (it == end())
		return false;
	*val_p = it->second->seq();
	return true;
}

bool DataSet::getSafeStr(std::string* val_p, Tag tag, VR vr) const
{
	const_iterator it = find(tag);
	if (it == end())
		return false;
	*val_p = it->second->str(vr);
	return true;
}

bool DataSet::getSafeStr(std::string* val_p, Tag tag, VR vr, int index) const
{
	const_iterator it = find(tag);
	if (it == end())
		return false;
	*val_p = it->second->str(vr, index);
	return true;
}

bool DataSet::getSafeUS(UINT16* val_p, Tag tag, int index) const
{
	const_iterator it = find(tag);
	return it != end() && it->second->getSafeUS(val_p, index);
}

bool DataSet::getSafeSS(INT16* val_p, Tag tag, int index) const
{
	const_iterator it = find(tag);
	return it != end() && it->second->getSafeSS(val_p, index);
}

bool DataSet::getSafeUL(UINT32* val_p, Tag tag, int index) const
{
	const_iterator it = find(tag);
	return it != end() && it->second->getSafeUL(val_p, index);
}

bool DataSet::getSafeSL(INT32* val_p, Tag tag, int index) const
{
	const_iterator it = find(tag);
	return it != end() && it->second->getSafeSL(val_p, index);
}

bool DataSet::getSafeFL(float* val_p, Tag tag, int index) const
{
	const_iterator it = find(tag);
	return it != end() && it->second->getSafeFL(val_p, index);
}

bool DataSet::getSafeFD(double* val_p, Tag tag, int index) const
{
	const_iterator it = find(tag);
	return it != end() && it->second->getSafeFD(val_p, index);
}
