/****************************************************************************
          Copyright (C) 1998, Karl Franzens University, Graz

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND THE KARL FRANZENS
          UNIVERSITY DOES NOT MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          owned by the Karl Franzens University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include this copyright notice.
****************************************************************************/

/***************************************************************************
 *
 * Karl Franzens University, Graz
 * UCDMC99 DICOM Network Transport Libraries 
 * Version 0.1 Beta
 *
 * Technical Contact: gunter.zeilinger@kfunigraz.ac.at
 *
 ***************************************************************************/

#include "base.hpp"

#include "DcmFileSet.h"
#include "Decoder2.h"
#include "Encoder2.h"

using std::string;
using std::sort;
static const char* TYPECODE[] =
{
   "PATIENT",
   "STUDY",
   "SERIES",
   "IMAGE",
   "OVERLAY",
   "MODALITY LUT",
   "VOI LUT",
   "CURVE",
   "TOPIC",
   "VISIT",
   "RESULTS",
   "INTERPRETATION",
   "STUDY COMPONENT",
   "PRINT QUEUE",
   "FILM SESSION",
   "FILM BOX",
   "IMAGE BOX",
   "PRIVATE",
   "MRDR"
};

static Tag PATIENT_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_PAT_NAME,
   TAG_PAT_ID,
   TAG_NULL
};

static Tag STUDY_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_STUDY_DATE,
   TAG_STUDY_TIME,
   TAG_ACCESS_NO,
   TAG_STUDY_DESC,
   TAG_STUDY_INST_UID,
   TAG_STUDY_ID,
   TAG_NULL
};

static Tag SERIES_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_MODALITY,
   TAG_SERIES_INST_UID,
   TAG_SERIES_NO,
   TAG_NULL
};

static Tag IMAGE_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_IMAGE_NO,
   TAG_NULL
};

static Tag OVERLAY_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_OVERLAY_NO,
   TAG_NULL
};

static Tag MODALITY_LUT_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_LUT_NO,
   TAG_NULL
};

static Tag VOI_LUT_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_LUT_NO,
   TAG_NULL
};

static Tag CURVE_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_CURVE_NO,
   TAG_NULL
};

static Tag TOPIC_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_TOPIC_TITLE,
   TAG_TOPIC_SUBJECT,
   TAG_TOPIC_AUTHOR,
   TAG_TOPIC_KEYWORDS,
   TAG_NULL
};

static Tag VISIT_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_INSTITUT_NAME,
   TAG_ADMISSION_ID,
   TAG_ADMIT_DATE,
   TAG_NULL
};

static Tag RESULTS_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_INST_CREATE_DATE,
   TAG_RESULT_ID,
   TAG_NULL
};

static Tag INTERPRETATION_KEYS[] =
{
   TAG_CHAR_SET,
	TAG_INTERPRET_TRANS_DATE,
	TAG_INTERPRET_AUTHOR,
	TAG_INTERPRET_DIAG_DESC,
	TAG_INTERPRET_DIAG_CODE_SEQ,
	TAG_INTERPRET_ID,
	TAG_INTERPRET_TYPE_ID,
	TAG_INTERPRET_STATUS_ID,
   TAG_NULL
};

static Tag STUDY_COMPONENT_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_MODALITY,
   TAG_STUDY_DESC,
   TAG_PROC_CODE_SEQ,
   TAG_PERF_PHYS_NAME,
   TAG_NULL
};

static Tag PRINT_QUEUE_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_PRINTER_NAME,
   TAG_PRINT_QUEUE_ID,
   TAG_NULL
};

static Tag FILM_SESSION_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_FILM_SESSION_LABEL,
   TAG_EXECUTION_STATUS,
   TAG_NULL
};

static Tag FILM_BOX_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_EXECUTION_STATUS,
   TAG_NULL
};

static Tag IMAGE_BOX_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_IMAGE_POS,
   TAG_NULL
};

static Tag PRIVATE_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_NULL
};

static Tag MRDR_KEYS[] =
{
   TAG_CHAR_SET,
   TAG_NULL
};

static Tag* KEYS[] =
{
   PATIENT_KEYS,
   STUDY_KEYS,
   SERIES_KEYS,
   IMAGE_KEYS,
   OVERLAY_KEYS,
   MODALITY_LUT_KEYS,
   VOI_LUT_KEYS,
   CURVE_KEYS,
   TOPIC_KEYS,
   VISIT_KEYS,
   RESULTS_KEYS,
   INTERPRETATION_KEYS,
   STUDY_COMPONENT_KEYS,
   PRINT_QUEUE_KEYS,
   FILM_SESSION_KEYS,
   FILM_BOX_KEYS,
   IMAGE_BOX_KEYS,
   PRIVATE_KEYS,
   MRDR_KEYS
};

void CreateDirRec::make(DataSet* pOut, Type type)
{
   pOut->clear();
   pOut->insertElem(TAG_OFFSET_NEXT_REC, UINT32(0));
   pOut->insertElem(TAG_REC_IN_USE, UINT16(0xffff));
   pOut->insertElem(TAG_OFFSET_REF_LOWER, UINT32(0));
   pOut->insertElem(TAG_REC_TYPE, VR_CS, TYPECODE[type]);
}

void CreateDirRec::makeFrom(DataSet* pOut, Type type, const DataSet& src)
{
   make(pOut, type);
   pOut->insertElemFrom(KEYS[type], src);
}

void CreateDirRec::addRefFile(DataSet* pOut, const string& fileID, const FileMetaInfo& fmi)
{
   pOut->insertElem(TAG_REF_FILE_ID, VR_CS, fileID);
   pOut->insertElem(TAG_REF_FILE_SOP_CLASS_UID, VR_UI,
      fmi.getStr(TAG_MEDIA_SOP_CLASS_UID, VR_UI));
   pOut->insertElem(TAG_REF_FILE_SOP_INST_UID, VR_UI,
      fmi.getStr(TAG_MEDIA_SOP_INST_UID, VR_UI));
   pOut->insertElem(TAG_REF_FILE_TS_UID, VR_UI,
      fmi.getStr(TAG_TRANSFER_SYNTAX_UID, VR_UI));
}

void CreateDirRec::makePatKeyFrom(DataSet* pOut, const DataSet& src) const
{
   makeFrom(pOut, PATIENT, src);
}

void CreateDirRec::makeStudyKeyFrom(DataSet* pOut, const DataSet& src) const
{
   makeFrom(pOut, STUDY, src);
}

void CreateDirRec::makeSeriesKeyFrom(DataSet* pOut, const DataSet& src) const
{
   makeFrom(pOut, SERIES, src);
}

void CreateDirRec::makeImageKeyFrom(DataSet* pOut, const DataSet& src) const
{
   makeFrom(pOut, IMAGE, src);
}

DcmFileSet::DcmFileSet()
{
   init();
}

void DcmFileSet::init()
{
   m_consistency =
      m_firstRecPosPos = m_lastRecPosPos = m_consistencyPos = m_sqLenPos =
      m_firstRecPos = m_lastRecPos = m_sqLen = m_newRecPos = 0;
}

DcmFileSet::DcmFileSet(const string& id, const string& instUID, TS ts)
: m_id(id), m_fmi(UID_MEDIA_DIR_STORAGE_SOP_CLASS, instUID, ts)
{
   init();
}

DcmFileSet::~DcmFileSet()
{

}

bool DcmFileSet::writeFileSetInfo()
{
	if (!m_fmi.writeTo(*this))
		return false;

   TS ts = m_fmi.getTS();
	Encoder2 encoder(*this, ts);

   encoder.write(DataSet::value_type(TAG_FILE_SET_ID, new Value(VR_CS, m_id)));
   if (!m_descFile.empty())
      encoder.write(DataSet::value_type(TAG_DESC_FILE_ID, new Value(VR_CS, m_descFile)));
   if (!m_charSet.empty())
      encoder.write(DataSet::value_type(TAG_CHAR_SET_DESC_FILE, new Value(VR_CS, m_charSet)));
   encoder.write(DataSet::value_type(TAG_OFFSET_FIRST_REC, new Value(m_firstRecPos)));
   m_firstRecPosPos = (UINT32)tellp() - 4;
   encoder.write(DataSet::value_type(TAG_OFFSET_LAST_REC, new Value(m_lastRecPos)));
   m_lastRecPosPos = (UINT32)tellp() - 4;
   encoder.write(DataSet::value_type(TAG_FILE_SET_CONS, new Value(m_consistency)));
   m_consistencyPos = (UINT32)tellp() - 2;
   encoder.writeTag(TAG_REC_SEQ);
   if (ts.isExplicitVR())
   {
      put('S');
      put('Q');
      put(0);
      put(0);
   }
   m_sqLenPos = (UINT32)tellp();
   encoder.write32(m_sqLen = 0);
   m_newRecPos = (UINT32)tellp();
   return good();
}

bool DcmFileSet::readFileSetInfo()
{
	if (!m_fmi.readFrom(*this))
		return false;

   TS ts = m_fmi.getTS();
	Decoder2 decoder(*this, ts);

   if (!readFileSetInfo1(decoder))
      return false;

   if (!m_firstRecPosPos || !m_lastRecPosPos || !m_consistencyPos)
      return false;

   if (decoder.readTag() != TAG_REC_SEQ)
      return false;

   if (ts.isExplicitVR())
   {
      if (get() != 'S' ||
         get() != 'Q')
         return false;
      get();
      get();
   }
   m_sqLenPos = (UINT32)tellg();
   m_sqLen = decoder.read32();  
   if (m_firstRecPos && m_firstRecPos != (UINT32)tellg())
      return false;

   if (m_lastRecPos)
   {
      DataSet ds;
      if (!readItem(&ds, m_lastRecPos))
         return false;
   }
   m_newRecPos = (UINT32)tellg();

   return true;
}

bool DcmFileSet::readFileSetInfo1(Decoder2& decoder)
{
   DataSet::value_type elem;
   do
   {
   	decoder.read(elem);
      switch (elem.first)
      {
      case TAG_FILE_SET_ID:
         m_id = elem.second->str(VR_CS);
         break;
      case TAG_DESC_FILE_ID:
         m_descFile = elem.second->str(VR_CS);
         break;
      case TAG_CHAR_SET_DESC_FILE:
         m_charSet = elem.second->str(VR_CS);
         break;
      case TAG_OFFSET_FIRST_REC:
         m_firstRecPos = elem.second->getUL();
         m_firstRecPosPos = (UINT32)tellg() - 4;
         break;
      case TAG_OFFSET_LAST_REC:
         m_lastRecPos = elem.second->getUL();
         m_lastRecPosPos = (UINT32)tellg() - 4;
         break;
      case TAG_FILE_SET_CONS:
         m_consistency = elem.second->getUS();
         m_consistencyPos = (UINT32)tellg() - 2;
         return true;
      }
   } while (elem.first < TAG_REC_SEQ);

   return false;
}

void DcmFileSet::setConsistency(bool flag)
{
   DcmFileSet::write16(m_consistencyPos,
      m_consistency = flag ? (UINT16)0 : (UINT16)0xffff);
}

bool DcmFileSet::readItem(DataSet* pRec, UINT32 pos)
{
   if (!pos)
      return false;
   flush();
   seekg(pos);
   Decoder2 decoder(*this, m_fmi.getTS());
   Tag tag;
   decoder.readItem(*pRec, tag);
   return tag == TAG_ITEM;
}

static const int POS_NEXT = 16;
static const int POS_INUSE = 28;
static const int POS_LOWER = 38;

UINT32 DcmFileSet::next(UINT32 pos)
{
   flush();
   return read32(pos + POS_NEXT); 
}

UINT32 DcmFileSet::lower(UINT32 pos)
{
   flush();
   return read32(pos + POS_LOWER); 
}

bool DcmFileSet::inUse(UINT32 pos)
{
   flush();
   return read16(pos + POS_INUSE) != 0; 
}

UINT32 DcmFileSet::writeItem(const DataSet& ds)
{
   seekp(m_lastRecPos = m_newRecPos);
   Encoder2 encoder(*this, m_fmi.getTS());
   encoder.writeItem(ds);
   m_newRecPos = (UINT32)tellp();
   if (m_sqLen == -1)
   {
	   encoder.writeTag(TAG_SEQ_DELIM_ITEM);
      encoder.write32(UINT32(0x00000000));
   }
   if (m_firstRecPos == 0)
   {
      seekp(m_firstRecPosPos);
      encoder.write32(m_firstRecPos = m_lastRecPos);
   }
   seekp(m_lastRecPosPos);
   encoder.write32(m_lastRecPos);
   if (m_sqLen != -1)
   {
      seekp(m_sqLenPos);
      encoder.write32(m_sqLen = m_newRecPos - m_sqLenPos - 4);
   }
   return m_lastRecPos;
}

void DcmFileSet::setNext(UINT32 pos, UINT32 val)
{
   UINT32 temp;
   if (temp = next(pos))
      setNext(temp, val);
   else
      write32(pos + POS_NEXT, val);
}

void DcmFileSet::setLower(UINT32 pos, UINT32 val)
{
   UINT32 temp;
   if (temp = lower(pos))
      setNext(temp, val);
   else
      write32(pos + POS_LOWER, val);
}

void DcmFileSet::setInUse(UINT32 pos)
{
   write16(pos + POS_INUSE, (UINT16)0xffff);
}

void DcmFileSet::setInActive(UINT32 pos)
{
   if (inUse(pos))
   {

      PosVec posVec(1, pos);
      getChildRecPos(posVec, pos);
      PosVec::iterator it = posVec.begin();
      PosVec::iterator end = posVec.end();
      sort(it, end);
      while (it != end)
         write16(*it++ + POS_INUSE, (UINT16)0);
   }
}

void DcmFileSet::getChildRecPos(PosVec& posVec, UINT32 pos, UINT16 deep)
{
   for (pos = lower(pos); pos; pos = next(pos))
   {
      if (inUse(pos))
      {
         posVec.push_back(pos);
         getChildRecPos(posVec, pos, deep-1);
      }
   }
}

UINT32 DcmFileSet::read32(UINT32 pos)
{
   seekg(pos);
   Decoder2 decoder(*this, m_fmi.getTS());
   return decoder.read32();
}

UINT16 DcmFileSet::read16(UINT32 pos)
{
   seekg(pos);
   Decoder2 decoder(*this, m_fmi.getTS());
   return decoder.read16();
}

void DcmFileSet::write32(UINT32 pos, UINT32 val)
{
   seekp(pos);
   Encoder2 encoder(*this, m_fmi.getTS());
   encoder.write32(val);
}

void DcmFileSet::write16(UINT32 pos, UINT16 val)
{
   seekp(pos);
   Encoder2 encoder(*this, m_fmi.getTS());
   encoder.write16(val);
}

UINT32 DcmFileSet::findPatKey(DataSet* pRec, UINT32 pos, const string& id)
{
   return findKey(pRec, pos, id, CreateDirRec::PATIENT, TAG_PAT_ID, VR_LO);
}

UINT32 DcmFileSet::findStudyKey(DataSet* pRec, UINT32 pos, const string& id)
{
   return findKey(pRec, pos, id, CreateDirRec::STUDY, TAG_STUDY_INST_UID, VR_UI);
}

UINT32 DcmFileSet::findSeriesKey(DataSet* pRec, UINT32 pos, const string& id)
{
   return findKey(pRec, pos, id, CreateDirRec::SERIES, TAG_SERIES_INST_UID, VR_UI);
}

UINT32 DcmFileSet::findImageKey(DataSet* pRec, UINT32 pos, const string& id)
{
   return findKey(pRec, pos, id, CreateDirRec::IMAGE, TAG_SOP_INST_UID, VR_UI);
}

UINT32 DcmFileSet::findKey(DataSet* pRec, UINT32 pos, const string& id, 
                                CreateDirRec::Type type, Tag tag, VR vr)
{
   string str;
   bool vr_ui = vr == VR_UI;
   while (pos)
   {
      if (!readItem(pRec, pos))
         return UINT32(-1);

      if (!pRec->getSafeStr(&str, TAG_REC_TYPE, VR_CS))
         return UINT32(-1);

      if (str != TYPECODE[type])
         continue;
      
      if (!(vr_ui && pRec->getSafeStr(&str, TAG_REF_FILE_SOP_INST_UID, VR_UI)) &&
         !pRec->getSafeStr(&str, tag, vr))
         return UINT32(-1);

      if (str == id)
         return pos;

      if (!pRec->getSafeUL(&pos,TAG_OFFSET_NEXT_REC))
         return UINT32(-1);
   }
   return pos;
}

UINT32 DcmFileSet::regImageFile(const string& path,
            const FileMetaInfo& fmi, const DataSet& ds, const CreateDirRec& cdr)
{
   DataSet temp;
   string patId, studyInstUid, seriesInstUid, sopInstUid;

   if (!ds.getSafeStr(&patId, TAG_PAT_ID, VR_LO)
      || !ds.getSafeStr(&studyInstUid, TAG_STUDY_INST_UID, VR_UI)
      || !ds.getSafeStr(&seriesInstUid, TAG_SERIES_INST_UID, VR_UI)
      || !ds.getSafeStr(&sopInstUid, TAG_SOP_INST_UID, VR_UI))
      return UINT32(-1);

   UINT32 root = m_firstRecPos;
   UINT32 patPos = findPatKey(&temp, root, patId);
   UINT32 studyPos = 0;
   switch (patPos)
   {
   case 0:
      cdr.makePatKeyFrom(&temp, ds);
      patPos = writeItem(temp);
      if (root != 0)
         setNext(root, patPos);
      break;
   case -1:
      return UINT32(-1);
   default:
       if (!temp.getSafeUL(&studyPos,TAG_OFFSET_REF_LOWER))
          return UINT32(-1);
   }
   UINT32 seriesPos = 0;
   switch (studyPos = findStudyKey(&temp, studyPos, studyInstUid))
   {
   case 0:
      cdr.makeStudyKeyFrom(&temp, ds);
      studyPos = writeItem(temp);
      setLower(patPos, studyPos);
      break;
   case -1:
      return UINT32(-1);
   default:
       if (!temp.getSafeUL(&seriesPos,TAG_OFFSET_REF_LOWER))
          return UINT32(-1);
   }
   UINT32 imagePos = 0;
   switch (seriesPos = findSeriesKey(&temp, seriesPos, seriesInstUid))
   {
   case 0:
      cdr.makeSeriesKeyFrom(&temp, ds);
      seriesPos = writeItem(temp);
      setLower(studyPos, seriesPos);
      break;
   case -1:
      return UINT32(-1);
   default:
       if (!temp.getSafeUL(&imagePos,TAG_OFFSET_REF_LOWER))
          return UINT32(-1);
   }
   switch (imagePos = findImageKey(&temp, imagePos, sopInstUid))
   {
   case 0:
      cdr.makeImageKeyFrom(&temp, ds);
      CreateDirRec::addRefFile(&temp, path, fmi);
      imagePos = writeItem(temp);
      setLower(seriesPos, imagePos);
      return 0;
   case -1:
      return UINT32(-1);
   }
   return imagePos;
}
