/****************************************************************************
          Copyright (C) 1998, Karl Franzens University, Graz

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND THE KARL FRANZENS
          UNIVERSITY DOES NOT MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          owned by the Karl Franzens University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include this copyright notice.
****************************************************************************/

/***************************************************************************
 *
 * Karl Franzens University, Graz
 * UCDMC99 DICOM Network Transport Libraries 
 * Version 0.1 Beta
 *
 * Technical Contact: gunter.zeilinger@kfunigraz.ac.at
 *
 ***************************************************************************/

#include "base.hpp"

#include "Decoder.h"
#include "DataDict.h"
#include "pdata.hpp"

#include "Endian.h"

UINT32 Decoder::read(DataSet::value_type& elem)
{

	Tag& tag = (Tag&)(elem.first);
	UINT32 len;

	m_buffer >> tag;
	elem.second.bind(0);

	switch (tag)
	{
	case TAG_ITEM_DELIM_ITEM:
		m_buffer >> len;
		if (len != 0)
			m_err = ERR_LEN0;
		return sizeof(UINT16) + sizeof(UINT16) + sizeof(UINT32);
	case TAG_ITEM:
		m_err = ERR_ITEMTAG;
		return sizeof(UINT16) + sizeof(UINT16);
	case TAG_SEQ_DELIM_ITEM:
		m_err = ERR_SEQ_DELTAG;
		return sizeof(UINT16) + sizeof(UINT16);
	}

	VR vr = VR_UN;
	UINT32 readLen = sizeof(UINT16) + sizeof(UINT16);
	if (m_ts.isExplicitVR())
	{
		UINT16 w;
		m_buffer >> w;
		vr = VR(w);
		m_buffer >> w;
		len = w;
		readLen += sizeof(UINT16) + sizeof(UINT16);
	}

	if (vr == VR_UN || vr == VR_SQ || vr == VR_OW || vr == VR_OB)
	{
		m_buffer >> len;
		readLen += sizeof(UINT32);
	}

	if (len != 0xffffffff && len > MAX_VALID_LENGTH)
	{
		m_err = ERR_LEN_INVALID;
		return readLen;
	}

	if (getGrTag(tag) != 0x0000 && !m_ts.isExplicitVR())
	{
		if (tag == TAG_PIXEL_DATA ||
			tag == TAG_OVERLAY_DATA ||
			tag == TAG_RED_PAL_LUT ||
			tag == TAG_GREEN_PAL_LUT ||
			tag == TAG_BLUE_PAL_LUT ||
			tag == TAG_SEG_RED_PAL_LUT ||
			tag == TAG_SEG_GREEN_PAL_LUT ||
			tag == TAG_SEG_BLUE_PAL_LUT)
			vr = VR_OW;
		else if (tag == TAG_CURVE_DATA)
			vr = VR_OB;
		else
		{
			DataDict::instance().lookup(tag, &vr);
		}
	}

	Value* pValue = new Value(vr);
	elem.second.bind(pValue);

	if (len == 0)
		return readLen;

	if (len == 0xffffffff || vr == VR_SQ)
	{
		Sequence* pSeq = new Sequence;
		pValue->setSQ(pSeq);
		return readLen + read(pSeq, len);
	}
	char* data_p = new char[len];
	pValue->setData(vr, len, data_p);
	switch(m_ts.isBigEndian() ? vr : VR_OB)
	{
	case VR_US:
	case VR_SS:
	case VR_AT:
	case VR_OW:
		{
			LE_UINT16* pData = (LE_UINT16*)data_p;
			BE_UINT16 temp;
			for (int count = len / sizeof(UINT16); count--;)
			{
				if (!m_buffer.Read((BYTE*)&temp, sizeof(UINT16)))
				{
					m_err = ERR_READ;
					return readLen;
				}
				*pData++ = temp;
				readLen += sizeof(UINT16);
			}
		}
		break;
	case VR_UL:
	case VR_SL:
	case VR_FL:
		{
			LE_UINT32* pData = (LE_UINT32*)data_p;
			BE_UINT32 temp;
			for (int count = len / sizeof(UINT32); count--;)
			{
				if (!m_buffer.Read((BYTE*)&temp, sizeof(UINT32)))
				{
					m_err = ERR_READ;
					return readLen;
				}
				*pData++ = temp;
				readLen += sizeof(UINT32);
			}
		}
		break;
	case VR_FD:
		{
			LE_double* pData = (LE_double*)data_p;
			BE_double temp;
			for (int count = len / sizeof(double); count--;)
			{
				if (!m_buffer.Read((BYTE*)&temp, sizeof(double)))
				{
					m_err = ERR_READ;
					return readLen;
				}
				*pData++ = temp;
				readLen += sizeof(double);
			}
		}
		break;
	default:
		{
			if (!m_buffer.Read((BYTE*)data_p, len))
			{
				m_err = ERR_READ;
				return readLen;
			}
			readLen += len;
		}
	}
	return readLen;
}

UINT32 Decoder::read(DataSet& ds, UINT32 len)
{
	ds.clear();
	UINT32 readLen = 0;
	while (m_err == NO_ERR && m_buffer.GetIncomingSize() > 0 && readLen < len)
	{
		DataSet::value_type elem;
		readLen += read(elem);

		switch(elem.first)
		{
		case TAG_ITEM_DELIM_ITEM:
		case TAG_DATA_SET_PADDING:
			return readLen;
		}
		if (getElTag(elem.first) != 0x0000)
			ds.insert(elem);
	}
	return readLen;
}

UINT32 Decoder::read(Sequence* pSeq, UINT32 len)
{
	UINT32 readLen = 0;

	while (m_err == NO_ERR && readLen < len)
	{
		Tag tag;
		UINT32 itemLen;

		m_buffer >> tag;
		m_buffer >> itemLen;
		readLen += 8;

		switch(tag)
		{
		case TAG_SEQ_DELIM_ITEM:
			if (itemLen != 0)
				m_err = ERR_LEN0;
			return readLen;
		case TAG_ITEM:
			break;
		default:
			m_err = ERR_NOITEMTAG;
			return readLen;
		}
		DataSet* pDS = new DataSet;
		readLen += read(*pDS, itemLen);
		pSeq->push_back(pDS);
	}
	return readLen;
}
