/****************************************************************************
          Copyright (C) 1998, Karl Franzens University, Graz

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND THE KARL FRANZENS
          UNIVERSITY DOES NOT MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          owned by the Karl Franzens University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include this copyright notice.
****************************************************************************/

/***************************************************************************
 *
 * Karl Franzens University, Graz
 * UCDMC99 DICOM Network Transport Libraries 
 * Version 0.1 Beta
 *
 * Technical Contact: gunter.zeilinger@kfunigraz.ac.at
 *
 ***************************************************************************/

#include "base.hpp"

#include "Dumper.h"
#include "DataDict.h"
#include "Endian.h"

using std::string;
const Dumper::Scheme Dumper::defaultScheme =
{
	'#',		// noPrintCh;		 
	64,			// truncChar;	 
	32,			// truncByte;	 
	16,			// truncWord;	 
	"..",		// truncSign;	 
	"%.2x ",	// fOB;		 
	"%.4x ",	// fOW;		 
	"%.4d ",	// fUS;		 
	"%.8d ",	// fUL;		 
	"%+d ",		// fSS;		 
	"%+d ",		// fSL;		 
	"%g ",		// fFL;		 
	"%g ",		// fFD;		 
	"(%.4x,%.4x)",	// fTag;		 
	",%d",		// fLen;		 
	",%c%c",	// fVR;		 
	",%s",		// fName;		 
	"?????????????",	// unknown;	 
	"<",		// prefixSQ;	 
	"\t",		// prefixVal;	 
	"\n",		// endOfElem;	 
	"(null)",	// valSQelem;	 
	VR_OB		// fVR_UNas;		 
};

Dumper::Dumper(std::ostream& os, const Scheme& scheme)
:	m_os(os), m_scheme(scheme), m_SQlevel(0)
{
}

void Dumper::write(const char* s)
{
	m_os << s;
}

void Dumper::write(const DataSet::value_type& elem)
{
	if (!m_scheme.prefixSQ && elem.second->seq())
	{
		write(elem.second->seq());
		return;
	}

	char buffer[64];
	for (int i = m_SQlevel; i--;)
		m_os << m_scheme.prefixSQ;
	sprintf(buffer, m_scheme.fTag, getGrTag(elem.first), getElTag(elem.first));
	m_os << buffer;
	const Value& val = *elem.second;
	VR vr = VR_UN;
	const char* name = m_scheme.unknown;
	DataDict::instance().lookup(elem.first, &vr, &name);
	if (val.vr() != VR_UN)
		vr = val.vr();

	if (m_scheme.fVR != 0)
	{
		sprintf(buffer, m_scheme.fVR, getCh1(vr), getCh2(vr));
		m_os << buffer;
	}
	UINT32 len = val.length();
	if (m_scheme.fLen != 0)
	{
		sprintf(buffer, m_scheme.fLen, len);
		m_os << buffer;
	}
	if (m_scheme.fName != 0)
	{
		sprintf(buffer, m_scheme.fName, name);
		m_os << buffer;
	}

	m_os << m_scheme.prefixVal;
	const Sequence* pSQ = val.seq();
	if (pSQ != 0)
	{
		++m_SQlevel;
		m_os << m_scheme.valSQelem << m_scheme.endOfElem;
		write(pSQ);
		--m_SQlevel;
		return;
	}

	bool trunc = false;
	switch (vr != VR_UN ? vr : m_scheme.fVR_UNas)
	{
	case VR_AT:
		{
			const UINT32* pData = (UINT32*)val.data();
			trunc = len/8 > m_scheme.truncWord;
			for(int k = trunc ? m_scheme.truncWord : len/8; k != 0; --k, ++pData)
			{
				sprintf(buffer, m_scheme.fTag,
					(UINT16)((LE_UINT16*)pData)[0],
					(UINT16)((LE_UINT16*)pData)[1]);
				m_os << buffer;
			}
		}
		break;
	case VR_FL:
		{
			const LE_float* pData = (LE_float*)val.data();
			trunc = len/4 > m_scheme.truncWord;
			for(int k = trunc ? m_scheme.truncWord : len/4; k != 0; --k, ++pData)
			{
				sprintf(buffer, m_scheme.fFL, (float)*pData);
				m_os << buffer;
			}
		}
		break;
	case VR_FD:
		{
			const LE_double* pData = (LE_double*)val.data();
			trunc = len/8 > m_scheme.truncWord;
			for(int k = trunc ? m_scheme.truncWord : len/8; k != 0; --k, ++pData)
			{
				sprintf(buffer, m_scheme.fFD, (double)*pData);
				m_os << buffer;
			}
		}
		break;
	case VR_OB:
		{
			const BYTE* pData = (BYTE*)val.data();
			trunc = len > m_scheme.truncByte;
			for(int k = trunc ? m_scheme.truncByte : len; k != 0; --k, ++pData)
			{
				sprintf(buffer, m_scheme.fOB, *pData);
				m_os << buffer;
			}
		}
		break;
	case VR_OW:
		{
			const LE_UINT16* pData = (LE_UINT16*)val.data();
			trunc = len/2 > m_scheme.truncWord;
			for(int k = trunc ? m_scheme.truncWord : len/2; k != 0; --k, ++pData)
			{
				sprintf(buffer, m_scheme.fOW, (UINT16)*pData);
				m_os << buffer;
			}
		}
		break;
	case VR_SL:
		{
			const LE_INT32* pData = (LE_INT32*)val.data();
			trunc = len/4 > m_scheme.truncWord;
			for(int k = trunc ? m_scheme.truncWord : len/4; k != 0; --k, ++pData)
			{
				sprintf(buffer, m_scheme.fSL, (INT32)*pData);
				m_os << buffer;
			}
		}
		break;
	case VR_SQ:
		m_os << m_scheme.valSQelem;
		break;
	case VR_SS:
		{
			const LE_INT16* pData = (LE_INT16*)val.data();
			trunc = len/2 > m_scheme.truncWord;
			for(int k = trunc ? m_scheme.truncWord : len/2; k != 0; --k, ++pData)
			{
				sprintf(buffer, m_scheme.fSS, (INT16)*pData);
				m_os << buffer;
			}
		}
		break;
	case VR_UL:
		{
			const LE_UINT32* pData = (LE_UINT32*)val.data();
			trunc = len/4 > m_scheme.truncWord;
			for(int k = trunc ? m_scheme.truncWord : len/4; k != 0; --k, ++pData)
			{
				sprintf(buffer, m_scheme.fUL, (UINT32)*pData);
				m_os << buffer;
			}
		}
		break;
	case VR_US:
		{
			const LE_UINT16* pData = (LE_UINT16*)val.data();
			trunc = len/2 > m_scheme.truncWord;
			for(int k = trunc ? m_scheme.truncWord : len/2; k != 0; --k, ++pData)
			{
				sprintf(buffer, m_scheme.fUS, (UINT16)*pData);
				m_os << buffer;
			}
		}
		break;
	default:
		{
			m_os << '\"';
			if (len > 0)
			{
				string temp((const char*)val.data(), min(len, m_scheme.truncChar));
				for (const char* pData = temp.c_str(); *pData != '\0'; ++pData)
					m_os << (isprint(*pData) != 0 ? *pData : m_scheme.noPrintCh);
				if (len > m_scheme.truncChar)
					m_os << m_scheme.truncSign;
			}
			m_os << '\"';
		}
	}
	if (trunc)
		m_os << m_scheme.truncSign;
	m_os << m_scheme.endOfElem;
}

void Dumper::write(const DataSet& ds)
{
	DataSet::const_iterator iter = ds.begin();
	DataSet::const_iterator last = ds.end();
	for (; iter != last; ++iter)
		write(*iter);
}

void Dumper::write(const Sequence* pSQ)
{
	Sequence::const_iterator iter = pSQ->begin();
	Sequence::const_iterator last = pSQ->end();
	for (; iter != last; ++iter)
		write(**iter);
}
