/****************************************************************************
          Copyright (C) 1998, Karl Franzens University, Graz

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND THE KARL FRANZENS
          UNIVERSITY DOES NOT MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          owned by the Karl Franzens University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include this copyright notice.
****************************************************************************/

/***************************************************************************
 *
 * Karl Franzens University, Graz
 * UCDMC99 DICOM Network Transport Libraries 
 * Version 0.1 Beta
 *
 * Technical Contact: gunter.zeilinger@kfunigraz.ac.at
 *
 ***************************************************************************/

#include "base.hpp"

#include "FileMetaInfo.h"
#include "Decoder2.h"
#include "Encoder2.h"
#include "Impl.h"

using std::string;
using std::istream;
using std::ostream;
using std::streampos;

const string FileMetaInfo::VERSION("\0\1",2);

FileMetaInfo::FileMetaInfo()
{
   memset(m_preamble, 0, sizeof(m_preamble));
}

FileMetaInfo::FileMetaInfo(const string& classUID, const string& instUID, TS ts)
 : m_ts(ts)
{
  memset(m_preamble, 0, sizeof(m_preamble));
  insertElem(TAG_FILE_INFO_VERS, VR_OB, VERSION);
  insertElem(TAG_MEDIA_SOP_CLASS_UID, VR_UI, classUID);
  insertElem(TAG_MEDIA_SOP_INST_UID, VR_UI, instUID);
  insertElem(TAG_TRANSFER_SYNTAX_UID, VR_UI, ts.getUID());
  insertElem(TAG_IMPL_CLASS_UID, VR_UI, IMPL_CLASS_UID);
  insertElem(TAG_IMPL_VERS_NAME, VR_SH, IMPL_VERSION_NAME);
}

static const char DICM[4] = { 'D','I','C','M' };

bool FileMetaInfo::isPart10(istream& is)
{
  char dicm[4];
  UINT32 oldPos = is.tellg();
  bool ret = is.seekg(128) && is.read(dicm, 4) && memcmp(dicm, DICM, 4) == 0;
  is.seekg(oldPos);
  return ret;
}


bool FileMetaInfo::readFrom(istream& is)
{
  char dicm[4];
  if (!is.seekg(0)
    || !is.read(m_preamble, sizeof(m_preamble))
    || !is.read(dicm, 4)
    || memcmp(dicm, DICM, 4) != 0)
  {
    is.seekg(0);
    return false;
  }

  DataSet::value_type elem;
  Decoder2 decoder(is, TS::EXPL_VR_LE);
  decoder.read(elem);
  Value& val = *elem.second;
  if (elem.first != TAG_FILE_INFO_GR_LEN || 
    val.vr() != VR_UL || val.length() != 4)
    return false;

  DataSet temp;
  temp.insert(elem);
  decoder.read(temp, val.getUL());
  string tsUid;
  if (!temp.getSafeStr(&tsUid, TAG_TRANSFER_SYNTAX_UID, VR_UI))
    return false;
  m_ts = TS(tsUid);
  swap(temp);

	
  return true;
}


bool FileMetaInfo::readACRFrom(istream &is)
{
  char dicm[4];
  if (!is.seekg(0)
    || !is.read(m_preamble, sizeof(m_preamble))
    || !is.read(dicm, 4)
    || memcmp(dicm, DICM, 4) != 0)
  {
    is.seekg(0);
    return false;
  }

  DataSet::value_type elem;
  Decoder2 decoder(is, TS::IMPL_VR_LE);
  decoder.read(elem);
  Value& val = *elem.second;
  if (elem.first != TAG_FILE_INFO_GR_LEN || 
    val.vr() != VR_UL || val.length() != 4)
    return false;

  DataSet temp;
  temp.insert(elem);
  decoder.read(temp, val.getUL());
  string tsUid;
  if (!temp.getSafeStr(&tsUid, TAG_TRANSFER_SYNTAX_UID, VR_UI))
    return false;
  m_ts = TS(tsUid);
  swap(temp);

  return true;
}

bool FileMetaInfo::verify() const
{
  DataSet::const_iterator pVersion = find(TAG_FILE_INFO_VERS);
  return
    pVersion != end() && 
    pVersion->second->length() == 2 &&
    (((BYTE*)pVersion->second->data())[1] & 0x01) &&
    exist(TAG_MEDIA_SOP_CLASS_UID) &&
    exist(TAG_MEDIA_SOP_INST_UID) &&
    exist(TAG_TRANSFER_SYNTAX_UID) &&
    exist(TAG_IMPL_CLASS_UID);
}

bool FileMetaInfo::writeTo(ostream& os) const
{
  ASSERT(os.tellp() == streampos(0));
  os.write(m_preamble, sizeof(m_preamble));
  os.write(DICM,4);

  Encoder2 encoder(os, TS::EXPL_VR_LE);
  encoder.write(*this);
  return true;
}
