/****************************************************************************
          Copyright (C) 1998, Karl Franzens University, Graz

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND THE KARL FRANZENS
          UNIVERSITY DOES NOT MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          owned by the Karl Franzens University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include this copyright notice.
****************************************************************************/

/***************************************************************************
 *
 * Karl Franzens University, Graz
 * UCDMC99 DICOM Network Transport Libraries 
 * Version 0.1 Beta
 *
 * Technical Contact: gunter.zeilinger@kfunigraz.ac.at
 *
 ***************************************************************************/

#include "base.hpp"

#include "Parser.h"
#include "DataDict.h"

#include "Endian.h"

const Parser::Scheme Parser::defaultScheme =
{
  64,		  // truncChar;	 
  32,		  // truncByte;	 
  16,		  // truncWord;	 
  "%hx %n",	  // fOB;		 
  "%hx %n",	  // fOW;		 
  "%hx %n",	  // fUS;		 
  "%x %n",	  // fUL;		 
  "%hd %n",	  // fSS;		 
  "%d %n",	  // fSL;		 
  "%g %n",	  // fFL;		 
  "%lg %n",	  // fFD;		 
  "(%hx,%hx)%n",  // fTag;		 
  ",%d%n",	  // fLen;		 
  ",%c%c%n",	  // fVR;		 
  "<",		  // prefixSQ;	 
  "\t",		  // prefixVal;	 
//"\n",		  // endOfElem;	 
};

Parser::Parser(std::istream& is, const Scheme& scheme)
:	m_is(is), m_scheme(scheme)
{
}

bool Parser::read(const char*& pBuffer, Tag& tag)
{
  UINT n;
  UINT16 gggg;
  UINT16 eeee;
  if (sscanf(pBuffer, m_scheme.fTag, &gggg, &eeee, &n) != 2)
    return false;
  tag = makeTag(gggg,eeee);
  pBuffer += n;
  return true;
}

bool Parser::read(DataSet::value_type& elem)
{
  char ch = m_is.peek();
  if (ch != *m_scheme.fTag && ch != *m_scheme.prefixSQ)
    return false;

  m_is.getline(m_buffer, sizeof(m_buffer)); 
//  m_is.ignore(strlen(m_scheme.endOfElem) - 1); 

  const char* pBuffer = m_buffer;
  const char* pStr = m_scheme.prefixSQ;
  std::string::size_type n0 = strlen(m_scheme.prefixSQ);
  UINT sqLevel = 0;
  while (strncmp(pBuffer, pStr, n0) == 0)
  {
    pBuffer += n0;
    ++sqLevel;
  }
  if (sqLevel > m_seqStack.size())
    return false;

  m_seqStack.resize(sqLevel);

  Tag& tag = (Tag&)(elem.first);
  if (!read(pBuffer, tag))
    return false;
  elem.second.bind(0);

  UINT32 n = 0;
  VR vr = VR_UN;
  if (m_scheme.fVR != 0)
  {
    char ch1;
    char ch2;
    if (sscanf(pBuffer, m_scheme.fVR, &ch1, &ch2, &n) != 2)
      return false;
    pBuffer += n;
    vr = makeVR(ch1, ch2);
  }
  if (vr == VR_UN) 
    DataDict::instance().lookup(elem.first, &vr);

  UINT32 len = 0;
  if (m_scheme.fLen != 0)
  {
    if (sscanf(pBuffer, m_scheme.fLen, &len, &n) != 1)
      return false;
    pBuffer += n;
  }
  Value* pValue = new Value(vr);
  elem.second.bind(pValue);

  if (len == 0)
    return true;

  if (vr == VR_SQ || len == 0xffffffff)
  {
    Sequence* pSeq = new Sequence;
    pValue->setSQ(pSeq);
    return true;
  }

  char* data_p = new char[(len += 1) &= ~1];	// ensure even len
  memset(data_p, vr == VR_OB || vr == VR_UN || vr == VR_UI ? '\0' : ' ', len);

  pValue->setData(vr, len, data_p);

  pBuffer = strstr(pBuffer, m_scheme.prefixVal);
  if (pBuffer == 0)
    return false;

  pBuffer += strlen(m_scheme.prefixVal);

  switch(vr)
  {
  case VR_AT:
    {
      LE_UINT16* pData = (LE_UINT16*)data_p;
      UINT16 gggg, eeee;
      for(UINT32 n = len / 8; n != 0; --n)
      {
	if (sscanf(pBuffer, m_scheme.fTag, &gggg, &eeee, &n) != 2)
		return false;
	*pData++ = gggg;
	*pData++ = eeee;
	pBuffer += n;
      }
    }
    break;
  case VR_FL:
    {
      LE_float* pData = (LE_float*)data_p;
      float temp;
      for(UINT32 n = len / 4; n != 0; --n)
      {
        if (sscanf(pBuffer, m_scheme.fFL, &temp, &n) != 1)
          return false;
        *pData++ = temp;
        pBuffer += n;
      }
    }
    break;
  case VR_FD:
    {
      LE_double* pData = (LE_double*)data_p;
      double temp;
      for(UINT32 n = len / 8; n != 0; --n, ++pData)
      {
        if (sscanf(pBuffer, m_scheme.fFD, &temp, &n) != 1)
          return false;
        *pData++ = temp;
        pBuffer += n;
      }
    }
    break;
  case VR_OB:
  case VR_UN:
    {
      UINT16 temp;
      BYTE* pData = (BYTE*)data_p;
      bool trunc = len > m_scheme.truncByte;
      UINT32 n = trunc ? m_scheme.truncByte : len;
      for(; n != 0; --n, ++pData)
      {
        if (sscanf(pBuffer, m_scheme.fOB, &temp, &n) != 1)
          return false;
        *pData++ = BYTE(temp);
        pBuffer += n;
      }
    }
    break;
  case VR_OW:
    {
      LE_UINT16* pData = (LE_UINT16*)data_p;
      UINT16 temp;
      bool trunc = len/2 > m_scheme.truncWord;
      UINT32 k = trunc ? m_scheme.truncWord : len/2;
      for(; k != 0; --k)
      {
        if (sscanf(pBuffer, m_scheme.fOW, &temp, &n) != 1)
          return false;
        *pData++ = temp;
        pBuffer += n;
      }
    }
    break;
  case VR_SL:
    {
      LE_INT32* pData = (LE_INT32*)data_p;
      INT32 temp;
      for(UINT32 k = len / 4; k != 0; --k)
      {
        if (sscanf(pBuffer, m_scheme.fSL, &temp, &n) != 1)
          return false;
        *pData++ = temp;
        pBuffer += n;
      }
    }
    break;
  case VR_SQ:
    ASSERT(false);
    break;
  case VR_SS:
    {
      LE_INT16* pData = (LE_INT16*)data_p;
      INT16 temp;
      for(UINT32 k = len / 2; k != 0; --k)
      {
        if (sscanf(pBuffer, m_scheme.fSS, &temp, &n) != 1)
          return false;
        *pData++ = temp;
        pBuffer += n;
      }
    }
    break;
  case VR_UL:
    {
      LE_UINT32* pData = (LE_UINT32*)data_p;
      UINT32 temp;
      for(UINT32 k = len / 4; k != 0; --k)
      {
        if (sscanf(pBuffer, m_scheme.fUL, &temp, &n) != 1)
          return false;
        *pData++ = temp;
        pBuffer += n;
      }
    }
    break;
  case VR_US:
    {
      LE_UINT16* pData = (LE_UINT16*)data_p;
      UINT16 temp;
      for(UINT32 k = len / 2; k != 0; --k)
      {
        if (sscanf(pBuffer, m_scheme.fUS, &temp, &n) != 1)
          return false;
        *pData++ = temp;
        pBuffer += n;
      }
    }
    break;
  default:
    {
      UINT k = strlen(pBuffer);
      if (k < 2 || pBuffer[0] != '\"' || pBuffer[k-1] != '\"')
        return false;

      if ((k -= 2) > len) k = len;
      if (k > m_scheme.truncChar) k = m_scheme.truncChar;
      memcpy(data_p, pBuffer+1, k);
    }
  }
  return true;
}

UINT Parser::read(DataSet& ds, UINT nMax)
{
  UINT n = 0;
  ds.clear();
  while (m_is.good() && n < nMax)
  {
    DataSet::value_type elem;
    if (!read(elem))
      return n;
    if (!m_seqStack.empty())
    {
      Sequence* pSQ = m_seqStack.back();
      HDataSet hDS;
      if (pSQ->empty() || elem.first <= (hDS = pSQ->back())->rbegin()->first)
      {
        pSQ->push_back(hDS = HDataSet(new DataSet));
      }
      hDS->insert(elem);
    } else
      ds.insert(elem);

    Sequence* pSQ = (Sequence*)elem.second->seq();
    if (pSQ != 0)
      m_seqStack.push_back(pSQ);
    ++n;
  }
  return n;
}

