/****************************************************************************
          Copyright (C) 1998, Karl Franzens University, Graz

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND THE KARL FRANZENS
          UNIVERSITY DOES NOT MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          owned by the Karl Franzens University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include this copyright notice.
****************************************************************************/

/***************************************************************************
 *
 * Karl Franzens University, Graz
 * UCDMC99 DICOM Network Transport Libraries 
 * Version 0.1 Beta
 *
 * Technical Contact: gunter.zeilinger@kfunigraz.ac.at
 *
 ***************************************************************************/

#include "base.hpp"

#include "Server.h"
#include "Pdu.h"
#include "MsgLogger.h"

#ifdef WIN32
#ifdef _MT
#include <process.h>
#endif
#endif
#ifdef SUNOS5
#include <thread.h>
#define _MT
#define __cdecl
#endif
#ifdef __VMS
#include <pthread.h>
#define _MT
#endif

Server::Server()
{
	bStopping = false;
  m_timeOut = 60;
  m_pAcceptRemoteApTitles = 0;
  m_pAcceptLocalApTitles = 0;
  m_pAcceptApplicationContexts = 0;
  m_acceptAbstractSyntaxs.insert(UID_VERIFICATION_SOP_CLASS);
}

void Server::handle(int sock)
{
  MsgLogger* msgLog_p = MsgLogger::get();
  ErrLogger* errLog_p = ErrLogger::get();
  PDU_Service pdu;
  pdu.Link.SetTimeOut(m_timeOut);
  pdu.SetAcceptRemoteApTitles(m_pAcceptRemoteApTitles);
  pdu.SetAcceptLocalApTitles(m_pAcceptLocalApTitles);
  pdu.SetAcceptApplicationContexts(m_pAcceptApplicationContexts);
  pdu.SetAcceptAbstractSyntaxs(&m_acceptAbstractSyntaxs);
  if (pdu.Multiplex(sock))
  {
    if (msgLog_p)
      msgLog_p->logAcceptAssoc(sock);
    DataSet command;		
    while (pdu.Read(command))
    {
      if (!m_dispatcher.dispatch(pdu, command))
      {
	if (errLog_p)
          errLog_p->logDispatchFailed(sock);
      }
    }
    switch (pdu.lastPdu())
    {
    case PDU_Service::AR_RQ:
    case PDU_Service::AAB_RQ:
      if (msgLog_p)
        msgLog_p->logReceivedPdu(sock, pdu.lastPdu());
      break;
    case PDU_Service::NO_PDU:
      if (errLog_p)
        errLog_p->logProtocolError(sock);
      break;
    case PDU_Service::TIMEOUT:
      if (errLog_p)
        errLog_p->logTimeOutError(sock);
    case PDU_Service::TCPCLOSE:
      if (errLog_p)
        errLog_p->logTcpCloseError(sock);
      break;
    default:
      if (errLog_p)
        errLog_p->logReceivedPdu(sock, pdu.lastPdu());
    }
    if (msgLog_p)
      msgLog_p->logReleaseAssoc(sock);
    onCloseAssoc(sock, pdu.lastPdu());
  } else
      if (msgLog_p)
        msgLog_p->logRejectAssoc(sock,
          pdu.AAssociateRJ::Result,
          pdu.AAssociateRJ::Source,
          pdu.AAssociateRJ::Reason);
}

void Server::onCloseAssoc(int sock, int lastPdu)
{
}

Server::ErrorCode Server::run(const std::string& port)
{
  Socket link;
  ErrLogger* errLog_p = ErrLogger::get();
  MsgLogger* msgLog_p = MsgLogger::get();
  if (!link.Listen(port.c_str()))
  {
    if (errLog_p)
      errLog_p->logListenSockFailed(port);
    return ERR_LISTEN;
  }
  if (msgLog_p)
    msgLog_p->logListenSock(port);
  while(link.Accept())
  {
	  if ( bStopping ) return ERR_NO;
    if (msgLog_p)
      msgLog_p->logAcceptSock(link.Socketfd);
    handle(link.Socketfd);
  }
  if (errLog_p)
    errLog_p->logAcceptSockFailed(port);
  return ERR_ACCEPT;
}

#ifdef _MT
typedef std::pair<Server*,int> ThreadParam;

#ifdef WIN32
void __cdecl start_func(void* p)
{
  MsgLogger* msgLog_p = MsgLogger::get();
  ThreadParam* param_p = (ThreadParam*)p;
  if (msgLog_p)
    msgLog_p->logStartThread(param_p->second);
  param_p->first->handle(param_p->second);
  if (msgLog_p)
    msgLog_p->logExitThread(param_p->second);
  delete param_p;
}

static bool startThread(ThreadParam* param_p)
{
  return _beginthread(start_func, 0, (void*)param_p) != -1;
}
#endif	// WIN32

#ifdef SUNOS5
void* start_func(void* p)
{
  MsgLogger* msgLog_p = MsgLogger::get();
  ThreadParam* param_p = (ThreadParam*)p;
  if (msgLog_p)
    msgLog_p->logStartThread(param_p->second);
  param_p->first->handle(param_p->second);
  if (msgLog_p)
    msgLog_p->logExitThread(param_p->second);
  delete param_p;
  return 0;
}

static bool startThread(ThreadParam* param_p)
{
  thread_t tid;
  return thr_create(NULL, 0, start_func, (void*)param_p, 0, &tid) == 0;
}
#endif	// SUNOS5

#ifdef __VMS
void* start_func(void* p)
{
  MsgLogger* msgLog_p = MsgLogger::get();
  ThreadParam* param_p = (ThreadParam*)p;
  if (msgLog_p)
    msgLog_p->logStartThread(param_p->second);
  param_p->first->handle(param_p->second);
  if (msgLog_p)
    msgLog_p->logExitThread(param_p->second);
  delete param_p;
  return 0;
}

static bool startThread(ThreadParam* param_p)
{
  pthread_t tid;
  return pthread_create(&tid, NULL, start_func, (void*)param_p) == 0;
}
#endif	// __VMS


Server::ErrorCode Server::runMultiThread(const std::string& port, bool abortIfErrThread)
{
  Socket link;
  ErrLogger* errLog_p = ErrLogger::get();
  MsgLogger* msgLog_p = MsgLogger::get();
  if (!link.Listen(port.c_str()))
  {
    if (errLog_p)
	    errLog_p->logListenSockFailed(port);
    return ERR_LISTEN;
  }
  if (msgLog_p)
    msgLog_p->logListenSock(port);
  while(link.Accept())
  {
	  if ( bStopping ) return ERR_NO;
    if (msgLog_p)
      msgLog_p->logAcceptSock(link.Socketfd);
    ThreadParam* param_p = new ThreadParam(this, link.Socketfd);
    if (!startThread(param_p))
    {
      if (errLog_p)
        errLog_p->logStartThreadFailed(param_p->second);
      delete param_p;
      if (abortIfErrThread)
        return ERR_THREAD;
    } else
    {
      link.Connected = 0;	
      link.Socketfd = 0;	
    }
  }
  if (errLog_p)
    errLog_p->logAcceptSockFailed(port);
  return ERR_ACCEPT;
}
#endif	// _MT
